"use strict";
exports.__esModule = true;
function ownKeys(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) {
    var symbols = Object.getOwnPropertySymbols(object);
    if (enumerableOnly)
        symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; });
    keys.push.apply(keys, symbols);
} return keys; }
function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};
    if (i % 2) {
        ownKeys(source, true).forEach(function (key) { _defineProperty(target, key, source[key]); });
    }
    else if (Object.getOwnPropertyDescriptors) {
        Object.defineProperties(target, Object.getOwnPropertyDescriptors(source));
    }
    else {
        ownKeys(source).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); });
    }
} return target; }
function _defineProperty(obj, key, value) { if (key in obj) {
    Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true });
}
else {
    obj[key] = value;
} return obj; }
// DOM Utils
var getElementClientRect_1 = require("./dom-utils/getElementClientRect");
var listScrollParents_1 = require("./dom-utils/listScrollParents");
var getWindow_1 = require("./dom-utils/getWindow");
var getNodeScroll_1 = require("./dom-utils/getNodeScroll");
var getOffsetParent_1 = require("./dom-utils/getOffsetParent");
var getCommonTotalScroll_1 = require("./dom-utils/getCommonTotalScroll"); // Pure Utils
var unwrapJqueryElement_1 = require("./utils/unwrapJqueryElement");
var orderModifiers_1 = require("./utils/orderModifiers");
var expandEventListeners_1 = require("./utils/expandEventListeners");
var computeOffsets_1 = require("./utils/computeOffsets");
var debounce_1 = require("./utils/debounce");
var validateModifiers_1 = require("./utils/validateModifiers"); // Default modifiers
var modifiers = require("./modifiers/index");
var defaultModifiers = Object.values(modifiers);
var INVALID_ELEMENT_ERROR = 'Invalid `%s` argument provided to Popper.js, it must be either a valid DOM element or a jQuery-wrapped DOM element, you provided `%s`';
var areValidElements = function (a, b) { return a instanceof Element && b instanceof Element; };
var defaultOptions = {
    placement: 'bottom',
    eventListeners: {
        scroll: true,
        resize: true
    },
    modifiers: [],
    strategy: 'absolute'
};
var Popper = /** @class */ (function () {
    function Popper(reference, popper, options) {
        if (options === void 0) { options = defaultOptions; }
        var _this = this;
        _defineProperty(this, "state", {
            placement: 'bottom',
            orderedModifiers: [],
            options: defaultOptions,
            modifiersData: {}
        });
        _defineProperty(this, "update", debounce_1["default"](function () { return new Promise(function (success, reject) {
            _this.forceUpdate();
            success(_this.state);
        }); }));
        // Unwrap `reference` and `popper` elements in case they are
        // wrapped by jQuery, otherwise consume them as is
        this.state.elements = {
            reference: unwrapJqueryElement_1["default"](reference),
            popper: unwrapJqueryElement_1["default"](popper)
        };
        var _a = this.state.elements, referenceElement = _a.reference, popperElement = _a.popper; // Store options into state
        this.state.options = _objectSpread({}, defaultOptions, {}, options); // Cache the placement in cache to make it available to the modifiers
        // modifiers will modify this one (rather than the one in options)
        this.state.placement = options.placement; // Don't proceed if `reference` or `popper` are invalid elements
        if (!areValidElements(referenceElement, popperElement)) {
            return;
        }
        this.state.scrollParents = {
            reference: listScrollParents_1["default"](referenceElement),
            popper: listScrollParents_1["default"](popperElement)
        }; // Order `options.modifiers` so that the dependencies are fulfilled
        // once the modifiers are executed
        this.state.orderedModifiers = orderModifiers_1["default"](defaultModifiers.concat(this.state.options.modifiers)) // Apply user defined preferences to modifiers
            .map(function (modifier) { return _objectSpread({}, modifier, {}, _this.state.options.modifiers.find(function (_a) {
            var name = _a.name;
            return name === modifier.name;
        })); }); // Validate the provided modifiers so that the consumer will get warned
        // of one of the custom modifiers is invalid for any reason
        if (__DEV__) {
            validateModifiers_1["default"](this.state.options.modifiers);
        } // Modifiers have the opportunity to execute some arbitrary code before
        // the first update cycle is ran, the order of execution will be the same
        // defined by the modifier dependencies directive.
        // The `onLoad` function may add or alter the options of themselves
        this.state.orderedModifiers.forEach(function (_a) {
            var onLoad = _a.onLoad, enabled = _a.enabled;
            return enabled && onLoad && onLoad(_this.state);
        });
        this.update().then(function () {
            // After the first update completed, enable the event listeners
            _this.enableEventListeners(_this.state.options.eventListeners);
        });
    } // Async and optimistically optimized update
    // it will not be executed if not necessary
    // debounced, so that it only runs at most once-per-tick
    // Syncronous and forcefully executed update
    // it will always be executed even if not necessary, usually NOT needed
    // use Popper#update instead
    Popper.prototype.forceUpdate = function () {
        var _a = this.state.elements, referenceElement = _a.reference, popperElement = _a.popper; // Don't proceed if `reference` or `popper` are not valid elements anymore
        if (!areValidElements(referenceElement, popperElement)) {
            return;
        } // Get initial measurements
        // these are going to be used to compute the initial popper offsets
        // and as cache for any modifier that needs them later
        this.state.measures = {
            reference: getElementClientRect_1["default"](referenceElement),
            popper: getElementClientRect_1["default"](popperElement)
        }; // Get scrollTop and scrollLeft of the offsetParent
        // this will be used in the `computeOffsets` function to properly
        // position the popper taking in account the scroll position
        // FIXME: right now we only look for a single offsetParent (the popper one)
        //        but we really want to take in account the reference offsetParent as well
        var offsetParentScroll = getNodeScroll_1["default"](getOffsetParent_1["default"](popperElement)); // Offsets are the actual position the popper needs to have to be
        // properly positioned near its reference element
        // This is the most basic placement, and will be adjusted by
        // the modifiers in the next step
        this.state.offsets = {
            popper: computeOffsets_1["default"]({
                reference: this.state.measures.reference,
                element: this.state.measures.popper,
                strategy: 'absolute',
                placement: this.state.placement,
                scroll: getCommonTotalScroll_1["default"](referenceElement, this.state.scrollParents.reference, this.state.scrollParents.popper)
            })
        }; // Modifiers have the ability to read the current Popper.js state, included
        // the popper offsets, and modify it to address specifc cases
        this.state.reset = false;
        for (var index = 0; index < this.state.orderedModifiers.length; index++) {
            var _b = this.state.orderedModifiers[index], fn = _b.fn, enabled = _b.enabled, options = _b.options;
            if (this.state.reset === true) {
                this.state.reset = false;
                index = 0;
                continue;
            }
            if (enabled && typeof fn === 'function') {
                this.state = fn(this.state, options);
            }
        }
    };
    Popper.prototype.enableEventListeners = function (eventListeners) {
        var _this = this;
        var _a = this.state.elements, referenceElement = _a.reference, popperElement = _a.popper;
        var _b = expandEventListeners_1["default"](eventListeners), scroll = _b.scroll, resize = _b.resize;
        if (scroll) {
            var scrollParents = this.state.scrollParents.reference.concat(this.state.scrollParents.popper);
            scrollParents.forEach(function (scrollParent) { return scrollParent.addEventListener('scroll', _this.update, {
                passive: true
            }); });
        }
        if (resize) {
            var window_1 = getWindow_1["default"](this.state.elements.popper);
            window_1.addEventListener('resize', this.update, {
                passive: true
            });
        }
    };
    return Popper;
}());
exports["default"] = Popper;
