"use strict";
exports.__esModule = true;
// This modifier takes the Popper.js state and prepares some StyleSheet properties
// that can be applied to the popper element to make it render in the expected position.
exports.mapStrategyToPosition = function (strategy) {
    switch (strategy) {
        case 'fixed':
            return 'fixed';
        case 'absolute':
        default:
            return 'absolute';
    }
};
exports.computePopperStyles = function (_a) {
    var offsets = _a.offsets, strategy = _a.strategy, gpuAcceleration = _a.gpuAcceleration;
    // by default it is active, disable it only if explicitly set to false
    if (gpuAcceleration === false) {
        return {
            top: offsets.y + "px",
            left: offsets.x + "px",
            position: exports.mapStrategyToPosition(strategy)
        };
    }
    else {
        return {
            transform: "translate3d(" + offsets.x + "px, " + offsets.y + "px, 0)",
            position: exports.mapStrategyToPosition(strategy)
        };
    }
};
exports.computeArrowStyles = function (_a) {
    var offsets = _a.offsets, gpuAcceleration = _a.gpuAcceleration;
    if (gpuAcceleration) {
        return {
            top: offsets.y + "px",
            left: offsets.x + "px",
            position: 'absolute'
        };
    }
    else {
        return {
            transform: "translate3d(" + offsets.x + "px, " + offsets.y + "px, 0)",
            position: 'absolute'
        };
    }
};
function computeStyles(state, options) {
    var gpuAcceleration = options && options.gpuAcceleration != null ? options.gpuAcceleration : true;
    state.styles = {}; // popper offsets are always available
    state.styles.popper = exports.computePopperStyles({
        offsets: state.offsets.popper,
        strategy: state.options.strategy,
        gpuAcceleration: gpuAcceleration
    }); // arrow offsets may not be available
    if (state.offsets.arrow != null) {
        state.styles.arrow = exports.computeArrowStyles({
            offsets: state.offsets.arrow,
            gpuAcceleration: gpuAcceleration
        });
    }
    return state;
}
exports.computeStyles = computeStyles;
exports["default"] = {
    name: 'computeStyles',
    enabled: true,
    phase: 'afterMain',
    fn: computeStyles
};
